// Example sketch showing how to create a simple addressed, reliable messaging client
// with the RadioHead library and driver to send messages over LoRa to our gateway.
// It works with our gateway example lora_server.
// Tested with Dragino Gateway & Arduino Uno with Dragino LoRa shield and BME280 sensor.
//
// This sketch requires the latest versions of the RadioHead and SparkFunBME280 libraries
// to be installed.
//
// Made by David Sluga <david.sluga@senlab.io> for
// SenLab d.o.o.
//

#include <RHReliableDatagram.h>
#include <RH_RF95.h>
#include <SPI.h>
#include <Wire.h>
#include <SparkFunBME280.h>

// Set the address of the gateway and node. It will also be set in the gateway sketch
// so it receives correct data from the nodes.
#define NODE_ADDRESS 2
#define GATEWAY_ADDRESS 1

// Set the reference sea level pressure for the BME280 sensor
#define SEALEVELPRESSURE_HPA (1013.25)

//Global sensor object
BME280 mySensor;

// Single instance of the radio driver
RH_RF95 driver;

// Class to manage message delivery and receipt, using the driver declared above
RHReliableDatagram manager(driver, NODE_ADDRESS);

// Required on Arduino Zero (or compatible) that uses SerialUSB instead of Serial
//#define Serial SerialUSB

// Set the radio frequency. Check allowed values at https://lpwanmarket.com/global-frequencies/
float frequency = 868.0;

void setup() {
	// Initialize the Serial so we can see diagnostic data
	Serial.begin(9600);
	// Wait for the Serial to become available to catch all output
	while (!Serial)
		;
	// Initialize the LoRa shield
	if (!manager.init())
		Serial.println(F("init failed"));
	// Defaults after init are 434.0MHz, 13dBm, Bw = 125 kHz, Cr = 4/5, Sf = 128chips/symbol, CRC on
	driver.setFrequency(frequency);
	driver.setModemConfig(RH_RF95::Bw125Cr45Sf128);
	// The default transmitter power is 13dBm, using PA_BOOST.
	// If you are using RFM95/96/97/98 modules which uses the PA_BOOST transmitter pin, then
	// you can set transmitter powers from 5 to 23 dBm:
	driver.setTxPower(14, false);
	// You can optionally require this module to wait until Channel Activity
	// Detection shows no activity on the channel before transmitting by setting
	// the CAD timeout to non-zero:
	//driver.setCADTimeout(10000);

	// Initialize the BME280 temperature, humidity, and pressure sensor
	// In this case it's connected to the I2C interface and located at address 0x76
	// Find more detailed descriptions of these options in examples from the SparkFun
	// BME 280 Arduino library
	mySensor.settings.commInterface = I2C_MODE;
	mySensor.settings.I2CAddress = 0x76;
	mySensor.settings.runMode = 3;
	mySensor.settings.tStandby = 0;
	mySensor.settings.filter = 0;
	mySensor.settings.tempOverSample = 1;
	mySensor.settings.pressOverSample = 1;
	mySensor.settings.humidOverSample = 1;

	//Make sure sensor had enough time to turn on. BME280 requires 2ms to start up.
	delay(10);
	//Calling .begin() causes the settings to be loaded
	mySensor.begin();
	Serial.print(F("Started BME280"));
}

// Declare the LoRa message variable, the size should be adjusted to reflect the number of
// expected characters in the outbound message + 1 for the null terminator
int8_t data[55];
// Declare the variable that stores the ACK reply sent by the gateway
uint8_t buf[RH_RF95_MAX_MESSAGE_LEN];

void loop() {

	// Declare and set variables for latitude and its sensor ID
	char latID = '1';
	char lat[9] = "46.03954";

	// Declare and set variables for longitude and its sensor ID
	char lonID = '2';
	char lon[9] = "14.47308";

	// Declare and set variables for temperature and its sensor ID, use dtostrf to save the
	// temperature provided as a float type into a char array
	char tempID = '3';
	char temp[6];
	dtostrf(mySensor.readTempC(), 0, 2, temp);

	// Declare and set variables for humidity and its sensor ID, use dtostrf to save the
	// temperature provided as a float type into a char array
	char humID='4';
	char hum[6];
	dtostrf(mySensor.readFloatHumidity(), 0, 2, hum);

	// Declare and set variables for pressure and its sensor ID, use dtostrf to save the
	// temperature provided as a float type into a char array
	char presID = '5';
	char pres[5];
	dtostrf(mySensor.readFloatPressure()/100, 0, 0, pres);

	// Assemble a char array delimited by semicolons, with every first character in a delimited
	// part representing the sensor ID and followed by the reading using the sprintf function
	sprintf((char*)data, "%c%s;%c%s;%c%s;%c%s;%c%s", latID, lat, lonID, lon, tempID, temp, humID, hum, presID, pres);

	// Print out the sensor data about to be sent
	Serial.println();
	Serial.print(F("Temperature: "));
	Serial.print(temp);
	Serial.println(F(" C"));
	Serial.print(F("Humidity:    "));
	Serial.print(hum);
	Serial.println(F(" %"));
	Serial.print(F("Pressure:    "));
	Serial.print(pres);
	Serial.println(F(" hPa"));

	// Print out the final message to be sent over LoRa
	Serial.println((char*)data);

	Serial.print(F("Sending to gateway 0x"));
	Serial.print(GATEWAY_ADDRESS, HEX);
	// Send the message to the gateway
	if (manager.sendtoWait((unsigned char*)data, sizeof(data), GATEWAY_ADDRESS)) {
		// Now wait for a reply from the gateway
		uint8_t len = sizeof(buf);
		uint8_t from;
		if (manager.recvfromAckTimeout(buf, &len, 2000, &from)) {
			Serial.print(F("Got reply from gateway 0x"));
			Serial.print(from, HEX);
			Serial.print(F(": "));
			Serial.println((char*) buf);
			Serial.print(F("RSSI: "));
			Serial.println(driver.lastRssi(), DEC);
			Serial.print(F("SNR: "));
			Serial.println(driver.lastSNR(), DEC);
			Serial.print(F("Frequency error: "));
			Serial.println(driver.frequencyError());
		} else {
			Serial.println(F("No reply, is the gateway running?"));
		}
	} else
		Serial.println(F("sendtoWait failed"));
	// Set the variable storing the received message to an empty char array
	data[0] = '\0';
	// Wait for the specified amount of time before the loop runs again
	delay(20000);
}

