// Example sketch showing how to create a simple addressed, reliable messaging server
// with the RadioHead library and driver to receive messages over LoRa and forward them
// to an MQTT server. It works with our other examples, like lora_client_bme280.
// Tested with Dragino Gateway & Arduino Uno with Dragino LoRa shield.
//
// This sketch requires the latest version of the RadioHead library to be installed.
//
// Made by David Sluga <david.sluga@senlab.io> for
// SenLab d.o.o.
//

#include <RHReliableDatagram.h>
#include <RH_RF95.h>
#include <SPI.h>
#include <Console.h>
#include <Bridge.h>

// Set the address of the gateway. It will also be used in the node sketches
// so they send their data to the correct server.
#define GATEWAY_ADDRESS 1

// Single instance of the radio driver
RH_RF95 driver;

// Class to manage message delivery and receipt, using the driver declared above
RHReliableDatagram manager(driver, GATEWAY_ADDRESS);

// Set the radio frequency. Check allowed values at https://lpwanmarket.com/global-frequencies/
float frequency = 868.0;

void setup() {
	// Initialize the Bridge, so the Arduino can communicate with the OpenWRT part
	Bridge.begin(115200);
	// Initialize the Console so we can see diagnostic data
	Console.begin();
	// Wait for the Console to become available to catch all output
	while (!Console)
		;
	// Initialize the LoRA shield
	if (!manager.init())
		Console.println(F("LoRa init failed"));
	// Defaults after init are 434.0MHz, 13dBm, Bw = 125 kHz, Cr = 4/5, Sf = 128chips/symbol, CRC on
	driver.setFrequency(frequency);
	driver.setModemConfig(RH_RF95::Bw125Cr45Sf128);
	// The default transmitter power is 13dBm, using PA_BOOST.
	// If you are using RFM95/96/97/98 modules which uses the PA_BOOST transmitter pin, then
	// you can set transmitter powers from 5 to 23 dBm:
	driver.setTxPower(14, false);
	// You can optionally require this module to wait until Channel Activity
	// Detection shows no activity on the channel before transmitting by setting
	// the CAD timeout to non-zero:
	//  driver.setCADTimeout(10000);
	Console.println(F("Start listening"));
}

uint8_t data[] = "ACK";
// Dont put this on the stack:
uint8_t buf[RH_RF95_MAX_MESSAGE_LEN];

void loop() {
	if (manager.available()) {
		// Wait for a message addressed to us from the client
		uint8_t len = sizeof(buf);
		uint8_t from;

		// Print out diagnostic info
		if (manager.recvfromAck(buf, &len, &from)) {
			Console.println();
			Console.print(F("Got message from node "));
			Console.println(from, HEX);
			Console.print("Message length: ");
			Console.println(len);
			Console.print(F("RSSI: "));
			Console.println(driver.lastRssi(), DEC);
			Console.print(F("SNR: "));
			Console.println(driver.lastSNR(), DEC);
			Console.print(F("Frequency error: "));
			Console.println(driver.frequencyError());

			// Send an acknowledgement back to the originating node
			if (!manager.sendtoWait(data, sizeof(data), from))
				Console.println(F("ACK sending failed"));

			// Print out the LoRa message
			Console.print(F("String: "));
			Console.println((char *)buf);

			// Save the time the message was received using the function getTimestamp()
			char *curTime = getTimestamp();

			char *message[len+1];
			int ii = -1;
			// Use strtok to split the received message by delimiter - semicolons in our case
			message[++ii] = strtok((char *)buf, ";");
			// Cut off the first character from the message since it represents the sensor ID
			char* reading = message[0] + 1;
			// Print out the information received in the first delimited part
			Console.print("Part ");
			Console.print(0);
			Console.print(": ");
			Console.println(message[0]);
			Console.print("ID: ");
			Console.print(message[0][0]);
			Console.print(", Reading: ");
			Console.println(reading);

			// Forward the first delimited part of the message to the mqtt_publish() function
			mqttPublish(from, message[0][0], reading, curTime);

			// Process the remaining delimited parts of the message using a while loop
			while ((message[++ii] = strtok(NULL, ";")) != NULL) {
				reading = message[ii] + 1;
				Console.print("Part ");
				Console.print(ii);
				Console.print(": ");
				Console.println(message[ii]);
				Console.print("ID: ");
				Console.print(message[ii][0]);
				Console.print(", Reading: ");
				Console.println(reading);
				mqttPublish(from, message[ii][0], reading, curTime);
			}
			Console.println();

			Console.print(F("Time stamp: "));
			Console.println(curTime);

			//Console.print(F("Before sending free RAM: "));
			//Console.println(freeRam());

			Console.println(F("Published to MQTT"));

			// Set the variable storing the received message to an empty char array
			buf[0] = '\0';
		}
	}
}

// freeRam() function returns the available quantity of RAM in bytes, useful for debugging
int freeRam() {
	extern int __heap_start, *__brkval;
	int v;
	return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}

// getTimestamp() function gets the current time in seconds since epoch form
// from the OpenWRT side of the gateway, and returns it as a char array
char * getTimestamp() {
	// Run the command "date +%s" to get seconds since epoch
	Process date;
	if (!date.running()) {
		date.begin(F("date"));
		date.addParameter(F("+%s"));
		date.run();
	}

	// Read the command output and save it to a char array
	static char timeDate[11];
	byte j = 0;
	char inChar;
	if (date.available() > 0) {
		while (date.available() > 0 && j < 10) {
			inChar = date.read();
			timeDate[j] = inChar;
			++j;
			timeDate[j] = '\0';
		}
	}

	return timeDate;
}

// mqttPublish() function takes the following data as input: ID of the sending node,
// ID of the sensor, value of the reading, and the timestamp
// It uses this data to call a bash script on the OpenWRT side of the gateway which
// does further processing and sends the data through MQTT to an IoTool server
void mqttPublish(uint8_t nodeid, char sensorid, char* reading,	char* timestamp) {
	// The addParameter method only takes char arrays as input so some data type
	// conversion is needed
	char node[4];
	sprintf(node, "%u", nodeid);
	char sensor[4];
	sprintf(sensor, "%c", sensorid);

    // Create the process that will call the bash script
	Process p;
	// Launch the "iotoolmqtt.sh" script
	p.begin(F("/root/iotoolmqtt.sh"));
	// Add the node ID
	p.addParameter(F("-n"));
	p.addParameter(node);
	// Add the sensor ID
	p.addParameter(F("-s"));
	p.addParameter(sensor);
	// Add the reading value
	p.addParameter(F("-r"));
	p.addParameter(reading);
	// Add the timestamp
	p.addParameter(F("-t"));
	p.addParameter(timestamp);
	// Enable MQTT debugging if needed
	p.addParameter(F("-d"));
	// Run the process and wait for its termination
	p.run();

	// The process output can be read with the stream methods
	while (p.available() > 0) {
		char c = p.read();
		Console.print(c);
	}
	// Ensure the last bit of data is sent.
	Console.flush();
}
